/******************************************************************************
 * Copyright (c) 2014, Hobu Inc. (howard@hobu.co)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided
 *       with the distribution.
 *     * Neither the name of the Howard Butler or Hobu, Inc.
 *       the names of its contributors may be
 *       used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 ****************************************************************************/
#pragma once

#include <iostream>
#include <stdlib.h>
#include <vector>

#include "Mathpair.hpp"

namespace hexer
{

class Hexagon;
class HexGrid;

class Segment
{
public:
    Segment() : m_hex(NULL), m_side(0)
        {}

    Segment(Hexagon *h, int side) : m_hex(h), m_side(side)
        {}

    Hexagon *hex()
        { return m_hex; }

    int side()
        { return m_side; }

    bool horizontal()
        { return (m_side == 0 || m_side == 3); }

    void normalize(HexGrid *grid);
    bool possibleRoot(HexGrid *grid);

    Segment rightAntiClockwise(HexGrid *grid);
    Segment leftAntiClockwise(HexGrid *grid);
    Segment rightClockwise(HexGrid *grid);
    Segment leftClockwise(HexGrid *grid);
    Point startPos(HexGrid *grid) const;
    Point endPos(HexGrid *grid) const;

    friend bool operator == (const Segment& s1, const Segment &s2);
    friend bool operator != (const Segment& s1, const Segment &s2);
    friend std::ostream& operator << (std::ostream& os, const Segment &s);

private:
    Point pos(HexGrid *grid, const Point& offset) const;

    /// Hexagon who's side is the segment.
    Hexagon *m_hex;
    /// Which side of the hexagon.
    int m_side;
};

} // namespace

